/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/database/low_bdb.c
 * @brief bdb based implementation of low database API
 *
 * @author Nils Durner
 **/

#include "config.h"

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <db.h>

#include "applications/afs/database/low_backend.h"
#include "util/xmalloc.h"
#include "util/logging.h"
#include "util/storage.h"
#include "util/returnvalues.h"
#include "util/statuscalls.h"
#include "applications/afs/encoding/block.h"

/**
 * After how-many insert operations test
 * DB size?
 **/
#define TEST_FREQUENCY 1024

/**
 * Extention for the Berkeley DB.
 **/
#define BDB_EXT ".bdb"

/**
 * Log DEBUG-Info?
 **/
#define BDB_DEBUG NO

/**
 * A bdb wrapper
 **/
typedef struct {

  /**
   * BDB handle.
   **/
  DB *dbf;

  /**
   * Number of insert operations since last
   * sync?
   **/
  int insertCount;

  int deleteSize;

  /**
   * Name of the database file.
   **/
  char * filename;

} bdbHandle;

/**
 * Open a bdb database (for content)
 * @param dir the directory where content is configured
 *         to be stored (e.g. data/content). A file 
 *         called ${dir}.bdb is used instead
 **/
static bdbHandle * getDatabase(char * dir)
{
  char * ff;
  bdbHandle * result;
  int ret;

  result = MALLOC(sizeof(bdbHandle));
  ff = MALLOC(strlen(dir)+strlen(BDB_EXT)+1);
  strcpy(ff, dir);
  if (ff[strlen(ff)-1] == '/')
    ff[strlen(ff) - 1] = 0; /* eat the '/' at the end */
  else 
    ff[strlen(ff)] = 0; /* no '/' to eat */
  strcat(ff, BDB_EXT); /* add the extention */
  result->filename = expandFileName(ff); /* expand ~/ if present */
  FREE(ff);

#if BDB_DEBUG
  LOG(LOG_DEBUG, 
      "BDB: Initializing the Berkeley DB\n");
#endif

  ret = db_create(&result->dbf, NULL, 0);
  if (ret != 0) 
  {
   LOG(LOG_ERROR, 
       "ERROR: Unable to init the Berkeley DB: %s",
       db_strerror(ret));

   goto err;
  }

  result->dbf->set_pagesize(result->dbf, 8192);

  // GNUnet accesses content randomly, so a mini cache is sufficient
  result->dbf->set_cachesize(result->dbf, 0, 20000, 0);

#if BDB_DEBUG
 LOG(LOG_DEBUG, "BDB: Opening datafile %s\n", result->filename);
#endif

#if DB_VERSION_MAJOR * 10 + DB_VERSION_MINOR >= 41
 ret = result->dbf->open(result->dbf, NULL, result->filename,
                         "data", DB_HASH, DB_CREATE | DB_THREAD,
                         S_IRUSR | S_IWUSR);
#else
  ret = result->dbf->open(result->dbf, result->filename,
                          "data", DB_HASH, DB_CREATE | DB_THREAD,
                          S_IRUSR | S_IWUSR);
#endif
  if (ret != 0) {
   LOG(LOG_ERROR, "ERROR: Unable to open the Berkeley DB: %s",
       db_strerror(ret));

   goto err;
  }

#if BDB_DEBUG
 LOG(LOG_DEBUG, 
     "BDB: Datafile opened\n");
#endif

  result->deleteSize = 0;

  return result;

  err:
   FREE(result);
   FREE(result->filename);
   return NULL;
}

void * lowInitContentDatabase(char * dir)
{  
  bdbHandle * dbh;
  
  dbh = getDatabase(dir);
  if (dbh == NULL) 
    errexit("FATAL: could not open database!\n");  
  return dbh;
}

/**
 * Clean shutdown of the storage module.
 **/
void lowDoneContentDatabase(void * handle)
{
  bdbHandle * dbh = handle;

#if BDB_DEBUG
 LOG(LOG_DEBUG, "BDB: Shutting down\n");
#endif
  
  dbh->dbf->close(dbh->dbf, 0);
  
  FREE(dbh->filename);
  FREE(dbh);
}

/**
 * Delete the BDB database.
 **/
void lowDeleteContentDatabase(void * handle)
{
  bdbHandle * dbh = handle;

#if BDB_DEBUG
 LOG(LOG_DEBUG, "BDB: Shutting down\n");
#endif
  
  dbh->dbf->close(dbh->dbf, 0);
  if (0 != remove(dbh->filename))
    LOG(LOG_ERROR,
	"ERROR: could not remove %s: %s\n",
	dbh->filename,
	strerror(errno));
  FREE(dbh->filename);
  FREE(dbh);
}

/**
 * Call a method for each entry in the database and
 * call the callback method on it. 
 * @return the number of items stored in the content database
 **/
int lowForEachEntryInDatabase(void * handle,
			      LowEntryCallback callback,
			      void * data)
{
  bdbHandle *dbh = handle;
  DBT key, dummy;
  HashCode160 doubleHash;
  int count = 0, ret;
  DBC *cursor;

  memset(&key, 0, sizeof(DBT));
  memset(&dummy, 0, sizeof(DBT));

#if BDB_DEBUG
 LOG(LOG_DEBUG, "BDB: Iterating through the database\n");
#endif
 
  /* scan database data/content.bdb and add entries to database 
     if not already present */
  dbh->dbf->cursor(dbh->dbf, NULL, &cursor, 0);

  while((ret = cursor->c_get(cursor, &key, &dummy, DB_NEXT)) != DB_NOTFOUND)
  {
   if (ret)
   {
    LOG(LOG_ERROR, "BDB: Unable to get next entry: %s\n",
       db_strerror(ret));
    
    break;
   }

#if BDB_DEBUG
 LOG(LOG_DEBUG, "BDB: Got next entry\n");
#endif

    if (ret == ENOMEM)
     continue;
    if ((key.size - 1) == sizeof(HashCode160)*2)
    {	
      if (callback != NULL)
      {
	hex2hash((HexName*)key.data,
		 &doubleHash);
	callback(&doubleHash, 
		 data);
      }
      count++; /* one more file */
    }
  }

#if BDB_DEBUG
 LOG(LOG_DEBUG, "BDB: Reached end of DB\n");
#endif

  cursor->c_close(cursor);

  return count;
}

#define COUNTENTRY "COUNT"

static void storeCount(bdbHandle * dbh,
		       int count)
{
  DBT key, buffer;

  memset(&key, 0, sizeof(DBT));
  memset(&buffer, 0, sizeof(DBT));
  key.data = COUNTENTRY;
  key.size = strlen(COUNTENTRY)+1;
  buffer.data = (char*)&count;
  buffer.size = sizeof(int);

#if BDB_DEBUG
 LOG(LOG_DEBUG, "BDB: Storing count\n");
#endif

  dbh->dbf->put(dbh->dbf, NULL, &key, &buffer, 0);
}

/**
 * Get the number of entries in the database.
 **/
int lowCountContentEntries(void * handle)
{
  bdbHandle * dbh = handle;
  DBT key, buffer;
  int count;

  memset(&key, 0, sizeof(DBT));
  memset(&buffer, 0, sizeof(DBT));
  key.data = COUNTENTRY;
  key.size = strlen(COUNTENTRY)+1;
  buffer.flags = DB_DBT_MALLOC;
  buffer.data = 0;

#if BDB_DEBUG
 LOG(LOG_DEBUG, "BDB: Retrieving count of entries\n");
#endif

  dbh->dbf->get(dbh->dbf, NULL, &key, &buffer, 0);
  if ( (!buffer.data) || 
       (buffer.size != sizeof(int)) )
  {
    count = lowForEachEntryInDatabase(dbh, NULL, NULL);  
    storeCount(dbh, count);
  } 
  else 
  {
    count = *(int*) buffer.data;
  }

  if (buffer.data)
   free(buffer.data);

  return count;
}

/**
 * Read the contents of a bucket to a buffer.
 *
 * @param name the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, SYSERR on failure
 **/ 
int lowReadContent(void * handle,
	     	   HashCode160 * name,
		   void ** result)
{
  bdbHandle * dbh = handle;
  HexName fn;
  DBT key, buffer;

  if ((name == NULL) || (result == NULL))
    return SYSERR;
  hash2hex(name, &fn);  
  
  memset(&key, 0, sizeof(DBT));
  memset(&buffer, 0, sizeof(DBT));
  key.data = fn.data;
  key.size = strlen(key.data) + 1;
  buffer.flags = DB_DBT_MALLOC;
  buffer.data = 0;

#if BDB_DEBUG
 LOG(LOG_DEBUG, "BDB: Retrieving data for key %s", key.data);
#endif

  dbh->dbf->get(dbh->dbf, NULL, &key, &buffer, 0);
  if (! buffer.data) 
    return SYSERR;
  *result = buffer.data;
  return buffer.size;  
}

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param name the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int lowWriteContent(void * handle,
		    HashCode160 * name, 
		    int len,
		    void * block)
{
  bdbHandle * dbh = handle;
  HexName fn;
  DBT buffer, key;

  hash2hex(name, &fn);
  
  memset(&key, 0, sizeof(DBT));
  memset(&buffer, 0, sizeof(DBT));
  key.data = fn.data;
  key.size = strlen(key.data) + 1;
  buffer.data = block;
  buffer.size = len;

#if BDB_DEBUG
 LOG(LOG_DEBUG, 
     "BDB: Storing data with the key %s", 
     key.data);
#endif

  dbh->dbf->put(dbh->dbf, NULL, &key, &buffer, 0);
  dbh->deleteSize -= len;
  if (dbh->deleteSize < 0)
    dbh->deleteSize = 0;
  storeCount(dbh, lowCountContentEntries(dbh) + 1);

  return OK;
}

/**
 * Free space in the database by removing one file
 * @param name hashcode representing the name of the file (without directory)
 * @return SYSERR on error, OK if ok.
 **/
int lowUnlinkFromDB(void * handle,
		    HashCode160 * name)
{
  bdbHandle *dbh = handle;
  DBT key, buffer;
  HexName fn;

  hash2hex(name, &fn);
  
  memset(&key, 0, sizeof(DBT));
  key.data = fn.data;
  key.size = strlen(key.data) + 1;

#ifdef BDB_DEBUG
  LOG(LOG_DEBUG, 
      "BDB: Deleting key %s", key.data);
#endif

  dbh->dbf->get(dbh->dbf, NULL, &key, &buffer, 0);
  dbh->dbf->del(dbh->dbf, NULL, &key, 0);
  if (buffer.data != NULL) {
    FREE(buffer.data);
    dbh->deleteSize += buffer.size;
  }
  storeCount(dbh, 
	     lowCountContentEntries(dbh) - 1);

  return OK;
}

/**
 * Estimate the size of the database.
 * @return the number of kb that the DB is assumed to use at the moment.
 **/
int lowEstimateSize(LowDBHandle handle) {
  bdbHandle * dbh = handle;

  return 
    ( (getFileSize(dbh->filename) * 120 / 100) -
      (dbh->deleteSize) +
      (sizeof(HashCode160) * lowCountContentEntries(handle))  
    ) / 1024; /* in kb */
}
 

/* end of low_bdb.c */
