/*
     This file is part of GNUnet.
     (C) 2001, 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/database/manager.c
 * @author Christian Grothoff
 *
 * This module is responsible to manage content, in particular 
 * it needs to decide what content to keep.
 **/

#include "config.h"

#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include "util/state.h"
#include "util/checksum.h"
#include "util/configuration.h"
#include "util/dso.h"
#include "util/statistics.h"
#include "util/logging.h"
#include "util/statuscalls.h"
#include "util/xmalloc.h"
#include "util/returnvalues.h"
#include "applications/afs/database/manager.h"
#include "applications/afs/database/high_backend.h"
#include "applications/afs/database/fileindex.h"
#include "applications/afs/database/bloomfilter.h"
#include "applications/afs/util/contentencoding.h"
#include "applications/afs/protocol/afs.h"
#include "applications/afs/protocol/messages.h"


#define DSO_PREFIX "libgnunetafs_database_"


/* ********************* GLOBALS ***************** */


/**
 * The current base value for fresh content (used to time-out old
 * content).
 **/
static int MANAGER_age;

/**
 * Is active migration allowed? This is about us receiving
 * data from the network, actively pushing content out is
 * always ok.
 **/
static int useActiveMigration;

/**
 * Global database handle
 **/
static DatabaseAPI * dbAPI = NULL;

/**
 * Statistics handles
 **/
static int stat_handle_lookup_3hash;
static int stat_handle_lookup_chk;
static int stat_handle_lookup_ondemand;
static int stat_handle_drivefull;

#define AGEFILE "database.age"

/**
 * Open the AGE file and return the handle.
 **/
static int getAgeFileHandle() {
  char * fileName;
  char * ef;
  int handle;
  
  LOG(LOG_CRON, 
      "CRON: enter cronReduceImportance\n");
  fileName = getFileName("AFS",
			 "AFSDIR",
			 "Configuration file must specify directory for"
			 " storage of AFS data in section %s under %s.\n");
  ef = MALLOC(strlen(fileName) + 
	      strlen(AGEFILE) +2);
  strcpy(ef, fileName);
  strcat(ef, "/");
  strcat(ef, AGEFILE);
  FREE(fileName);
  handle = open(ef,
		O_CREAT|O_RDWR,
		S_IRUSR|S_IWUSR);
  if (handle < 0) {
    LOG(LOG_WARNING,
	"Could not open agefile %s (%s)\n",
	ef,
	strerror(errno));
    FREE(ef);
    return SYSERR;
  }
  FREE(ef);
  return handle;
}

/**
 * Cron-job that decreases the importance-level of all
 * files by 1. Runs 'not very often'.
 **/
static void cronReduceImportance(void * unused) {
  int handle;
    
  LOG(LOG_CRON, 
      "CRON: enter cronReduceImportance\n");
  handle = getAgeFileHandle();
  if (handle == SYSERR)
    return;
  MANAGER_age++;
  write(handle, 
	&MANAGER_age,
	sizeof(int));
  close(handle);
  LOG(LOG_CRON, 
      "CRON: exit cronReduceImportance\n");
}

/**
 * Encode a block from a file on the drive, put the
 * result in the result buffer (allocate) and return
 * the size of the block.
 **/ 
static int encodeOnDemand(ContentIndex * ce,
			  CONTENT_Block ** result) {
  char * fn;
  int fileHandle;
  ssize_t blen;
  HashCode160 hc;
  CONTENT_Block * iobuf;
  HexName hex;

  /* on-demand encoding mechanism */
  fn = getIndexedFileName(ntohs(ce->fileNameIndex));
  if (fn == NULL) {
    LOG(LOG_FAILURE, 
	"FAILURE: Database inconsistent! "
	"(index points to invalid offset (%d)\n",
	ntohs(ce->fileNameIndex));
    return SYSERR;
  }
  fileHandle = open(fn, O_EXCL, S_IRUSR);
  if (fileHandle == -1) {
    LOG(LOG_FAILURE, 
	"FAILURE: Could not open file %s.\n",
	fn);   
    FREE(fn);
    return SYSERR;
  }
  FREE(fn);
  lseek(fileHandle, 
	ntohl(ce->fileOffset), 
	SEEK_SET);
  iobuf = MALLOC(sizeof(CONTENT_Block));
  blen = read(fileHandle, 
	      iobuf,
	      sizeof(CONTENT_Block));
  memset(&((char*)iobuf)[blen],
	 0,
	 sizeof(CONTENT_Block)-blen);
  /*LOG(LOG_DEBUG,
      "DEBUG: read %u bytes for on-demand encoding at %u\n",
      blen, 
      ntohl(ce->fileOffset));*/
  close(fileHandle);
  if (blen < 0) {
    LOG(LOG_ERROR,
	"ERROR: could not read file (%s)\n",
	strerror(errno));
    FREE(iobuf);
    return SYSERR;
  }
  hash(iobuf,
       blen, 
       &hc);
  *result = MALLOC(sizeof(CONTENT_Block));
  if (SYSERR == encryptContent(iobuf,
			       &hc,
			       *result)) 
    errexit("ERROR: encryption of content failed\n");

  FREE(iobuf);
  IFLOG(LOG_DEBUG,
	hash(*result,
	     sizeof(CONTENT_Block),
	     &hc);
	hash2hex(&hc,
		 &hex));
  /*LOG(LOG_DEBUG,
      "DEBUG: on-demand encoded content has query %s\n",
      &hex);*/
  return sizeof(CONTENT_Block);
}

/**
 * Load the high-level database as specified by
 * the given dtype.
 **/
DatabaseAPI * initializeDatabaseAPI(char * dtype) {
  DatabaseAPI * dbAPI;
  char * odtype;
  int len;
  void * lib;
  int i;

  if (dtype == NULL)
    errexit("AFS/DATABASETYPE not specified in config\n");
  odtype = NULL;
  len = stateReadContent("AFS-DATABASETYPE",
			 (void**)&odtype);
  if (len == -1) {
    stateWriteContent("AFS-DATABASETYPE",
		      strlen(dtype),
		      dtype);
  } else {
    if ( (len != strlen(dtype)) ||
	 (strncmp(dtype, odtype, len)) )
      errexit("FATAL: AFS database type was changed, run gnunet-convert\n");
  }
  FREENONNULL(odtype);
  dbAPI = MALLOC(sizeof(DatabaseAPI));
  lib = loadDynamicLibrary(DSO_PREFIX,
                           dtype);
  dbAPI->initContentDatabase    
    = bindDynamicMethod(lib,
			"",
			"initContentDatabase");
  dbAPI->doneContentDatabase
    = bindDynamicMethod(lib,
			"",
			"doneContentDatabase");
  dbAPI->forEachEntryInDatabase
    = bindDynamicMethod(lib,
			"",
			"forEachEntryInDatabase");
  dbAPI->countContentEntries 
    = bindDynamicMethod(lib,
			"",
			"countContentEntries");
  dbAPI->getMinimumPriority
    = bindDynamicMethod(lib,
			"",
			"getMinimumPriority");
  dbAPI->readContent    
    = bindDynamicMethod(lib,
			"",
			"readContent");
  dbAPI->writeContent   
    = bindDynamicMethod(lib,
			"",
			"writeContent");
  dbAPI->unlinkFromDB   
    = bindDynamicMethod(lib,
			"",
			"unlinkFromDB");
  dbAPI->deleteContent  
    = bindDynamicMethod(lib,
			"",
			"deleteContent");
  dbAPI->getRandomContent
    = bindDynamicMethod(lib,
			"",
			"getRandomContent");
  dbAPI->estimateAvailableBlocks
    = bindDynamicMethod(lib,
			"",
			"estimateAvailableBlocks");
  dbAPI->deleteDatabase
    = bindDynamicMethod(lib,
			"",
			"deleteDatabase");
  dbAPI->dynamicLibrary = lib;
  FREE(dtype);
  dbAPI->buckets
    = 4 * getConfigurationInt("AFS",
			      "DISKQUOTA") / 1024; /* one bucket per 250 MB */
  if (dbAPI->buckets == 0)
    dbAPI->buckets = 1; /* at least 1 bucket! */
  dbAPI->dbHandles
    = MALLOC(dbAPI->buckets * sizeof(HighDBHandle*));
  for (i=0;i<dbAPI->buckets;i++) {
    dbAPI->dbHandles[i] 
      = dbAPI->initContentDatabase(i, 
				   getConfigurationInt("AFS",
						       "DISKQUOTA"));
    if (dbAPI->dbHandles[i] == NULL)
      errexit("FATAL: failed to initialize AFS database %d\n", 
	      i);
  }
  return dbAPI;
}

typedef struct IterState {
  int hasNext;
  Semaphore * wsem;
  Semaphore * sem;
  HashCode160 next;
  ContentIndex ce;
  void * data;
  int len;
} IterState;

static void iterator_helper_callback(HashCode160 * key,
				     ContentIndex * ce,
				     void * data,
				     int dataLen,
				     IterState * state) {
  SEMAPHORE_DOWN(state->wsem);
  memcpy(&state->next,
	 key,
	 sizeof(HashCode160));
  memcpy(&state->ce,
	 ce,
	 sizeof(ContentIndex));
  state->data = data;
  state->len = dataLen;    
  SEMAPHORE_UP(state->sem);
}

/**
 * Thread that fetches the next entry from the database.
 * The thread is created by makeDatabaseIteratorState
 * and exits once we're through the database.
 **/
static void iterator_helper(IterState * state) {
  int i;

  state->hasNext = YES;
  for (i=0;i<dbAPI->buckets;i++)
    dbAPI->forEachEntryInDatabase(dbAPI->dbHandles[i],
				  (EntryCallback)&iterator_helper_callback,
				  state);
  SEMAPHORE_DOWN(state->wsem);
  state->hasNext = NO;
  SEMAPHORE_UP(state->sem);  
}

/**
 * Create the state required for a database iterator.  Calling this
 * method requires to call databaseIterator with the state returned
 * until "SYSERR" is returned.
 **/
void * makeDatabaseIteratorState() {
  IterState * ret;
  pthread_t pt;

  ret = MALLOC(sizeof(IterState));
  ret->sem = SEMAPHORE_NEW(0);
  ret->wsem = SEMAPHORE_NEW(1);
  if (0 != PTHREAD_CREATE(&pt,
			  (PThreadMain)&iterator_helper,
			  ret,
			  8 * 1024))
    errexit("could not create helper thread in %s:%d\n",
	    __FILE__, 
	    __LINE__);
  pthread_detach(pt);
  return ret;
}

/**
 * Iterator over all the queries in the database as needed by
 * resizeBloomfilter (and gnunet-check).  Typical use:
 * <code>
 * state = makeDatabaseIteratorState();
 * while (OK == databaseIterator(state, &x, &y, &z, &t))
 *   // do something
 * </code>
 * 
 * @param state the iterator state as created by
 *        makeDatabaseIteratorState
 * @returns OK if the iterator has filled in another element
 *  from the database, SYSERR if there are no more elements
 **/
int databaseIterator(void * state,
		     HashCode160 * hc,
		     ContentIndex * ce,
		     void ** data,
		     int * datalen) {
  IterState * st = state;

  SEMAPHORE_DOWN(st->sem);
  if (st->hasNext == NO) {
    SEMAPHORE_FREE(st->sem);
    SEMAPHORE_FREE(st->wsem);
    FREE(st);
    return SYSERR;
  }
  memcpy(hc, 
	 &st->next,
	 sizeof(HashCode160));
  memcpy(ce,
	 &st->ce,
	 sizeof(ContentIndex));
  *data = st->data;
  *datalen = st->len;
  SEMAPHORE_UP(st->wsem);
  return OK;
}

/**
 * Initialize the manager-module.
 **/
void initManager() {  
  int handle;
  int delta;
  int i;
  char * dtype;

  dtype = getConfigurationString("AFS",
  	 	 	         "DATABASETYPE");
  dbAPI = initializeDatabaseAPI(dtype);
  stat_handle_lookup_3hash
    = statHandle("# lookup (3HASH, search results)");
  stat_handle_lookup_chk
    = statHandle("# lookup (CHK, inserted or migrated content)");
  stat_handle_lookup_ondemand
    = statHandle("# lookup (ONDEMAND, indexed content)");
  stat_handle_drivefull
    = statHandle("# blocks AFS storage left (estimate)");
  handle = getAgeFileHandle();
  MANAGER_age = 0;
  if (handle != SYSERR) {
    read(handle, 
	 &MANAGER_age,
	 sizeof(int));
    close(handle);
  }
  useActiveMigration
    = testConfigurationString("AFS",
			      "ACTIVEMIGRATION",
			      "YES");
  addCronJob(&cronReduceImportance,
	     6 * cronHOURS,
	     12 * cronHOURS,
	     NULL); 

  delta = 0;
  for (i=0;i<dbAPI->buckets;i++)
    delta += dbAPI->estimateAvailableBlocks(dbAPI->dbHandles[i], 
					    getConfigurationInt("AFS",
								"DISKQUOTA")*1024/dbAPI->buckets);
  if (delta < 0) {
    for (i=0;i<dbAPI->buckets;i++) {
      dbAPI->deleteContent(dbAPI->dbHandles[i],
			   16-delta/dbAPI->buckets,
			   &bf_deleteEntryCallback,
			   NULL);
    }
    delta = 0;
  }
  statSet(stat_handle_drivefull,
	  delta);
}

/**
 * Shutdown the manager module.
 **/
void doneManager() {
  int i;

  delCronJob(&cronReduceImportance,
	     12 * cronHOURS,
	     NULL); 
  for (i=0;i<dbAPI->buckets;i++)
    dbAPI->doneContentDatabase(dbAPI->dbHandles[i]);
  FREE(dbAPI->dbHandles);
  unloadDynamicLibrary(dbAPI->dynamicLibrary);
  FREE(dbAPI);
  dbAPI = NULL;
}

/**
 * This function is as crazy as it is since RIPE160 hashes do not seem
 * to be quite random as we may want the to be...  So to get evenly
 * distributed indices, we have to be a bit tricky. And no, there is
 * high science but just a bit playing with the formula here.
 **/
static HighDBHandle * computeHighDB(HashCode160 * query) {
  HashCode160 qt;
  hash(query, sizeof(HashCode160), &qt);
  return dbAPI->dbHandles[( (unsigned int) (((query->a - qt.a) ^
					     (query->b - qt.b) ^ 
					     (query->c - qt.c) ^ 
					     (query->d - qt.d) ^
					     (query->e - qt.e))) >> 4) % dbAPI->buckets];
}

/**
 * Locate content. This method locates the data matching the
 * content entry. The data is on-demand encrypted if it is
 * indexed content or retrieved from the contentdatabase
 * if it was inserted content. 
 *
 * @param query the query for the content (CHK or 3HASH)
 * @param ce the meta-data for the content
 * @param result where to write the result, space will be
 *        allocated by retrieveContent, *result should be
 *        NULL when this function is invoked.
 * @return the length of the resulting content, SYSERR on error
 **/
int retrieveContent(HashCode160 * query,
		    ContentIndex * ce,
		    void ** result,
		    int prio) {
  int ret;
  
  ret = dbAPI->readContent(computeHighDB(query),
			   query,
      		  	   ce,
			   result,
			   prio);
  if (ret == -1) 
    return SYSERR;
  if (ntohs(ce->fileNameIndex) == 0) {
    switch (ntohs(ce->type)) {
    case LOOKUP_TYPE_CHK:
    case LOOKUP_TYPE_CHKS:
      statChange(stat_handle_lookup_chk, 1);
      break;
    case LOOKUP_TYPE_3HASH:
      statChange(stat_handle_lookup_3hash, 1);
      break;
    default:
      /* complain? */
    }
    return ret;
  } 
  if (*result != NULL) {
    LOG(LOG_ERROR,
	"ERROR: retrieved content but index says on-demand encoded!\n");
    FREE(*result);
    *result = NULL;
  }
  statChange(stat_handle_lookup_ondemand, 1);
  return encodeOnDemand(ce, 
			(CONTENT_Block**)result);
}
 
/**
 * Store content (if the priority is high enough), potentially
 * discarding less important content. 
 * 
 * @param ce the content entry describing the content
 * @param len the length of the data in bytes
 * @param data the block itself
 * @param sender from where does the content come? NULL for
 *        from local client.
 * @param duplicate output param, will be YES if content was already there
 * @return OK if the block was stored, SYSERR if not
 **/
int insertContent(ContentIndex * ce,
		  int len,
		  void * data,
		  HostIdentity * sender,
		  int * duplicate) {
  void * old;
  ContentIndex oldce;
  int oldLen;
  int avail;
  HashCode160 query;
  int i;
  int importance;
 
  *duplicate = NO;
  if ( (sender != NULL) &&
       (useActiveMigration == NO) )
    return SYSERR; /* forbidden! */
  importance = ntohl(ce->importance);
  if (importance < 0)
    importance = 0;
  if ( ( sender != NULL) &&
       ( randomi(2 + importance) == 0) )
    return SYSERR; /* don't bother... */
       
  ce->importance 
    = htonl(importance + MANAGER_age);
  if (ntohs(ce->type) == LOOKUP_TYPE_3HASH) {
    hash(&ce->hash, 
         sizeof(HashCode160),
         &query);
  } else {
    memcpy(&query,
	   &ce->hash,
	   sizeof(HashCode160));
  }

  memcpy(&oldce, 
	 ce, 
	 sizeof(ContentIndex));
  avail = 0;
  for (i=0;i<dbAPI->buckets;i++)
    avail += dbAPI->estimateAvailableBlocks(dbAPI->dbHandles[i],
					    getConfigurationInt("AFS",
								"DISKQUOTA")*1024/dbAPI->buckets);
  if (avail <= 0) {
    if (importance <= dbAPI->getMinimumPriority(computeHighDB(&query)))
      return SYSERR; /* new content has such a low priority that
			we should not even bother! */
    dbAPI->deleteContent(computeHighDB(&query),
			 16-avail,
			 &bf_deleteEntryCallback,
			 NULL);
    statSet(stat_handle_drivefull,
	    0);
  } else
    statSet(stat_handle_drivefull,
	    avail);

  /* try to read existing */
  old = NULL;
  oldLen = dbAPI->readContent(computeHighDB(&query),
			      &query,
			      &oldce,
			      &old,
			      0);
  /* add the content */
  if (ntohs(ce->type) == LOOKUP_TYPE_3HASH) {
				      
    /* multiple results are feasible, check! */
    if (oldLen != SYSERR) {
      int ret; 
      char * tmp;
      
      /* check if the content is already
	 present */
      for (ret=0;ret<oldLen/len;ret++) {
	if (0 == memcmp(&(((char*)old)[len*ret]),
			data,
			len)) {
	  /* content already there, abort */
	  FREENONNULL(old);
	  *duplicate = YES; 
	  return OK;
	}	   
      }
      tmp = MALLOC(oldLen + len);
      memcpy(tmp,
	     old,
	     oldLen);
      memcpy(&tmp[oldLen],
	     data,
	     len);
      /* FIXME: perhaps we should use eg max(a,b)? Otherwise n local inserts
         throws this through the ceiling... OTOH, if the same block was part
	 of two files, it is really twice as important, so adding makes sense
	 in some cases. */
      ce->importance = htonl(ntohl(oldce.importance) + ntohl(ce->importance));
      ret = dbAPI->writeContent(computeHighDB(&query),
				ce,
				oldLen + len,
				tmp);
      FREE(tmp);
      FREENONNULL(old);
      return ret;
    }
  } else {
    if (oldLen == len)
      *duplicate = YES;
  }
  FREENONNULL(old);
  if ( (*duplicate == NO) && 
       (SYSERR == dbAPI->writeContent(computeHighDB(&query),
				      ce,
				      len,
				      data)) ){
    LOG(LOG_DEBUG, 
	"DEBUG: writeContent() failed\n");
    return SYSERR; /* something went wrong... */
  }
  return OK;
}

/** 
 * Return a random key from the database.
 * @param ce output information about the key 
 * @return SYSERR on error, OK if ok.
 **/
int retrieveRandomContent(ContentIndex * ce) {
  int bucket = randomi(dbAPI->buckets);
  if (dbAPI->dbHandles[bucket] == NULL)
    errexit("dbHandle at bucket %d is NULL\n", 
	    bucket);
  return dbAPI->getRandomContent(dbAPI->dbHandles[bucket],
				 ce);
}

/**
 * Explicitly remove some content from the database.
 **/  
int removeContent(HashCode160 * query) {
  int ok;

  ok = dbAPI->unlinkFromDB(computeHighDB(query),
			   query);
  if (OK == ok) {
    int i;
    int delta = 0;
    for (i=0;i<dbAPI->buckets;i++)
      delta += dbAPI->estimateAvailableBlocks(dbAPI->dbHandles[i], 
					      getConfigurationInt("AFS",
								  "DISKQUOTA")*1024/dbAPI->buckets);
    if (delta < 0)  /* should not happen */
      delta = 0;
    statSet(stat_handle_drivefull,
	    delta);
  }
  return ok;
}

/* end of manager.c */
