/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/protocol/migration.c
 * @brief This module is responsible for pushing content out
 * into the network.
 * @author Christian Grothoff
 **/

#include "config.h"

#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include "util/checksum.h"
#include "util/statistics.h"
#include "util/logging.h"
#include "util/statuscalls.h"
#include "util/xmalloc.h"
#include "util/returnvalues.h"
#include "applications/afs/database/manager.h"
#include "applications/afs/protocol/afs.h"
#include "applications/afs/protocol/messages.h"
#include "applications/afs/protocol/migration.h"

static int stat_handle_content_pushed;

/**
 * Select content for active migration.
 **/
static int selectMigrationContent(HostIdentity * receiver,
				  ContentIndex * ce) {
  unsigned int avDist;
  unsigned int dist;
  int itc;  
  cron_t startTime;  
  int ok;
  
  cronTime(&startTime);
  itc = 0;
  avDist = 1;
  while (itc++ < 512) {
    if ( (cronTime(NULL) - startTime > 100 * cronMILLIS) ||
	 (getCPULoad() > 80) )
      return SYSERR; /* abort after 100ms or if load > 80 */
    ok = retrieveRandomContent(ce);
    if (ok == SYSERR)
      return SYSERR; /* failed */
    if (ntohs(ce->type) == LOOKUP_TYPE_3HASH ||
        ntohs(ce->type) == LOOKUP_TYPE_SUPER)
      continue; /* can not migrate these */
    
    dist = distanceHashCode160(&ce->hash,
			       &receiver->hashPubKey);
    /* division by 4 to ensure signed int range for randomi */ 
    if (randomi(1+dist/4) < randomi(1+avDist/4)) /* select */
      return OK;
    else
      avDist = (avDist * 31 + dist) / 32;
  }
  return SYSERR;
}
				  
/**
 * Build a CHK reply message for some content
 * selected for migration.
 * @return OK on success, SYSERR on error
 **/
static int buildCHKReply(ContentIndex * ce,
			 AFS_p2p_CHK_RESULT * pmsg) {
  CONTENT_Block * data;
  int ret;
  
  if (ntohs(ce->type) == LOOKUP_TYPE_3HASH ||
      ntohs(ce->type) == LOOKUP_TYPE_SUPER)
    return SYSERR;
  
  data = NULL;
  ret = retrieveContent(&ce->hash,
			ce,
			(void**)&data,
			0);
  if (ret == -1) /* can happen if we're concurrently inserting, 
		    _should be_ rare but is OK! */
    return SYSERR;
  if (ret != sizeof(CONTENT_Block)) {
    LOG(LOG_WARNING,
	"WARNING: buildCHKReply got unsuitable block from db (len=%d,type=%d)\n",
	ret, 
	ntohs(ce->type));
    FREENONNULL(data);
    return SYSERR;
  }
  pmsg->header.size 
    = htons(sizeof(AFS_p2p_CHK_RESULT));
  pmsg->header.requestType
    = htons(AFS_p2p_PROTO_CHK_RESULT);
  memcpy(&pmsg->result,
	 data,
	 sizeof(CONTENT_Block));
  FREE(data);
  return OK;
}

/**
 * Callback method for pushing content into the network.
 * The method chooses either a "recently" deleted block
 * or content that has a hash close to the receiver ID
 * (randomized to guarantee diversity, unpredictability
 * etc.).<p>
 *
 * @param receiver the receiver of the message
 * @param position is the reference to the
 *        first unused position in the buffer where GNUnet is building
 *        the message
 * @param padding is the number of bytes left in that buffer.
 * @return the number of bytes written to
 *   that buffer (must be a positive number).
 **/
static int activeMigrationCallback(HostIdentity * receiver,
				   char * position,
				   int padding) {
  AFS_p2p_CHK_RESULT * pmsg;
  int res;
  void * data;
  ContentIndex ce;
  
  res = 0;
  memset(&ce, 0, sizeof(ContentIndex));
  while (padding - res > sizeof(AFS_p2p_CHK_RESULT)) {
    data = NULL;
    if (SYSERR == selectMigrationContent(receiver,
					 &ce)) 
      return res; /* nothing selected, that's the end */
    /* append it! */
    pmsg = (AFS_p2p_CHK_RESULT*) &position[res];
    if (OK == buildCHKReply(&ce,
			    pmsg)) {
      statChange(stat_handle_content_pushed, 1);
      res += sizeof(AFS_p2p_CHK_RESULT);
    } else 
      return res; /* abort early after any error */    
  }
  return res;
}


void initMigration() {
  stat_handle_content_pushed
    = statHandle("# kb content pushed out as padding");
  coreAPI->registerSendCallback(sizeof(AFS_p2p_CHK_RESULT),
				(BufferFillCallback)&activeMigrationCallback);
}

void doneMigration() {
  coreAPI->unregisterSendCallback(sizeof(AFS_p2p_CHK_RESULT),
				  (BufferFillCallback)&activeMigrationCallback);
}

/* end of migration.c */
