/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Main function to insert multiple files into GNUnet.
 *
 * @author Christian Grothoff
 * @author James Blackwell
 * @file applications/afs/tools/gnunet-insert-multi.c 
 **/

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <sys/stat.h>
#include <dirent.h>

#include "util/checksum.h"
#include "util/configuration.h"
#include "util/getopt.h"
#include "util/initialize.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/xmalloc.h"
#include "applications/afs/afs.h"
#include "applications/afs/encoding/insertutil.h"
#include "applications/afs/encoding/keywords.h"
#include "applications/common/printhelp.h"
#include "util/port.h"


static void printstatus(ProgressStats * stats,
			void * verboselevel) {
  if (*((int*)verboselevel) == YES)
    printf("%8d of %8d bytes inserted\n",
	   stats->progress, stats->filesize);  
}

/**
 * Prints the usage information for this command if the user errs.
 * Aborts the program.
 **/
static void printhelp() {
  printf("USAGE: gnunet-insert-multi [OPTIONS] FILENAMES*\n"
         "-V, --verbose                 : "
         "be verbose\n"
	 "-p PRIORITY, --prio=PRIORITY  : "
	 "priority of the inserted content\n"
	 "-D DESC, --desc=DESC          : "
	 "force description for all files "
	 "(default: extract from files)\n"
	 "-k KEYWORD, --key=KEYWORD     : "
	 "additional keyword (can be specified multiple times)\n"
	 "-n, --noindex                 : "
	 "do not index, full insertion\n"
	 "-r, --recursive               : "
	 "index/insert subdirs recursively\n");
  printHelpDefault();
}

static int parseOptions(int argc,
			char ** argv) {
  int c;
  int num_keywords=0;
  char scratch[512];

  FREENONNULL(setConfigurationString("GNUNETD",
				     "LOGFILE",
				     NULL));
  FREENONNULL(setConfigurationString("GNUNET-INSERT",
				     "INDEX-CONTENT",
				     "YES"));
  FREENONNULL(setConfigurationString("GNUNET-INSERT",
				     "EXTRACT-KEYWORDS",
				     "YES"));
  while (1) {
    int option_index=0;
    static struct GNoption long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { "noindex",       0, 0, 'n' },
      { "prio",          1, 0, 'p' },
      { "key",           1, 0, 'k' },
      { "desc",          1, 0, 'D' },
      { "verbose",       0, 0, 'V' },
      { "recursive",     0, 0, 'r' },
      { 0,0,0,0 }
    };    
    c = GNgetopt_long(argc,
		      argv, 
		      "vVhdc:L:H:nD:k:p:r", 
		      long_options, 
		      &option_index);    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, optarg))
      continue;
    switch(c) {
    case 'D':
      FREENONNULL(setConfigurationString("GNUNET-INSERT",
					 "DESCRIPTION",
					 optarg));
      break;
    case 'k':
      sprintf(scratch, 
	      "KEYWORD.%d", 
	      num_keywords++);
      FREENONNULL(setConfigurationString("GNUNET-INSERT",
					 scratch,
					 optarg));
      break;
    case 'p': {
      unsigned int contentPriority;
      
      if (1 != sscanf(optarg,
		      "%ud", 
		      &contentPriority)) {
	LOG(LOG_FAILURE,
	    "FAILURE: You must pass a number to the -p option.\n");
	return SYSERR;
      }
      setConfigurationInt("GNUNET-INSERT",
			  "CONTENT-PRIORITY",
			  contentPriority);
      break;
    }
    case 'v': 
      printf("GNUnet v%s, gnunet-insert-multi v%s\n",
	     VERSION, 
	     AFS_VERSION);
      return SYSERR;
    case 'h': 
      printhelp(); 
      return SYSERR;
    case 'n':
      FREENONNULL(setConfigurationString("GNUNET-INSERT",
					 "INDEX-CONTENT",
					 "NO"));
      break;
    case 'V':
      FREENONNULL(setConfigurationString("GNUNET-INSERT",
                                         "VERBOSE",
                                         "YES"));
      break;
    case 'r':
      FREENONNULL(setConfigurationString("GNUNET-INSERT",
					 "RECURSIVE",
					 "YES"));
      break;
    default: 
      LOG(LOG_FAILURE,
	  "FAILURE: Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      return SYSERR;
    } /* end of parsing commandline */
  } /* while (1) */
  setConfigurationStringList(&argv[optind],
			     argc - optind);
  setConfigurationInt("GNUNET-INSERT",
   	  	      "KEYWORDCOUNT",
		      num_keywords);
  return OK;
}

/**
 * Insert a single file.
 * @param extractors list of extractors to use for keyword 
 *          extraction
 * @param sock the tcp socket to talk to gnunetd
 * @param filename the name of the file to insert
 * @returns OK on success, SYSERR on error
 **/
static int dofile(EXTRACTOR_ExtractorList * extractors,
		  GNUNET_TCP_SOCKET * sock,
		  char * filename) {
  int res = 0;
  int i;
  int num_keywords = 0;
  int verbose;
  char ** keywords = NULL;
  char * filenameRoot = NULL;
  char * mimetype;
  char * description;
  Block * top;

  if (YES == testConfigurationString("GNUNET-INSERT",
                                     "VERBOSE",
                                     "YES")) {
    verbose = YES;
  }

  if (YES == verbose)
    printf("Working on file %s...\n",
	   filename);
 
  /* do we have additional keys? */
  num_keywords=getConfigurationInt("GNUNET-INSERT",
                                   "KEYWORDCOUNT");
  if (num_keywords > 0) {
    char scratch[512];

    keywords = (char **)MALLOC(num_keywords*sizeof(char*));
    for (i=0;i<num_keywords;i++) {
      sprintf(scratch, "KEYWORD.%d", i);  
      keywords[i] = getConfigurationString("GNUNET-INSERT",
      					   scratch);
    }
  }
  description = getConfigurationString("GNUNET-INSERT",
				       "DESCRIPTION");
  mimetype = getConfigurationString("GNUNET-INSERT",
				    "MIMETYPE");
    
  extractKeywordsMulti(filename,
		       &description,
		       &mimetype,
		       &keywords,
		       &num_keywords,
		       extractors);
  if (num_keywords <= 0) {
    LOG(LOG_FAILURE,
	"Automatic keyword extraction failed, did not insert file %s.\n",
	filename);
    return -1;
  }  
  if (YES == testConfigurationString("GNUNET-INSERT",
				     "VERBOSE",
				     "YES")) {
    int k;

    printf("Using keywords:\n");
    for (k=0;k<num_keywords;k++)
      printf("%3d: %s\n",
	     k,
	     keywords[k]);
  }  
  if (description == NULL)
    description = STRDUP("No description supplied");
  if (mimetype == NULL)
    mimetype = STRDUP("unknown");
  
  /* if filename is '/home/user/foo', use 'foo' as the filenameRoot */
  filenameRoot = filename;
  for (res = 0;res<strlen(filename);res++)
    if (filename[res] == '/')
      filenameRoot = &filename[res+1];
  
  /* We finally get to do the real work! */
  top = insertFile(sock,
		   filename, 
		   &printstatus,
		   &verbose);
  if (top == NULL)
    res = SYSERR;
  else
    res = insertRoot(sock,
		     top,
		     description,
		     filenameRoot,
		     mimetype,
		     num_keywords,
		     keywords);   
  
  if (res == OK) {
    if(verbose == YES) {
      HexName hex;
      HexName hex2;
      printf("File %s successfully indexed ",
	     filename);
      hash2hex(&top->chk.key, 
	       &hex);
      hash2hex(&top->chk.query, 
	       &hex2);
      printf("-- %s %s %d %d\n", 
	     (char*)&hex, 
	     (char*)&hex2,
	     crc32N(top->data, top->len), 
	     top->filesize);
    }
  } else
    printf("Error inserting file %s.\n"
	   "You may want to check whether or not you are out of space.\n"
	   "Run gnunet-stats | grep \"AFS storage left\" to check.\n",
	   filename);
  if (top != NULL)
    top->done(top, NULL);
  while (num_keywords > 0) 
    FREE(keywords[--num_keywords]);
  FREE(mimetype);
  FREE(description);
  return res;
}

/**
 * Forward declaration
 **/
static int processObj(EXTRACTOR_ExtractorList *pExtractors,
		      GNUNET_TCP_SOCKET *pSock, char *pszObj,
		      int bRecursive);

/**
 * Recurse the passed directory and insert/index all files
 * @param pszDir the directory
 * @returns OK on success, SYSERR on error
 **/
static int recurseDir(EXTRACTOR_ExtractorList *pExtractors,
	              GNUNET_TCP_SOCKET *pSock, const char *pszDir,
		      int bRecurse)
{
 DIR *pDir;
 struct dirent *pEntry;
 int iRet;
 char *lpszObj = NULL;
 unsigned int uiAlloc = 0, uiLen, uiMem;

 pDir = opendir(pszDir);
 if (pDir)
 {
  iRet = OK;
  while ( (pEntry = readdir(pDir)) )
  {
   // Suppress "." and ".."
   if (!strcmp(pEntry->d_name, ".") || !strcmp(pEntry->d_name, ".."))
   {
    continue;
   }

   // Do we need to allocate additional memory?
   uiLen = strlen(pszDir);
   uiMem = uiLen + strlen(pEntry->d_name) + 2;
   if (uiAlloc < uiMem)
   {
    GROW(lpszObj,
          uiAlloc,
	  uiMem);
   }
   strcpy(lpszObj, pszDir);

   // Do we have to add a trailing slash?
   if (lpszObj[uiLen - 1] != '/')
   {
    lpszObj[uiLen++] = '/';
   }

   // Append the entryname
   strcpy(lpszObj + uiLen, pEntry->d_name);

   // Process it!
   if (processObj(pExtractors, pSock, lpszObj, bRecurse) == SYSERR)
   {
    iRet = SYSERR;
   }
  }
  closedir(pDir);
 }
 else
 {
  fprintf(stderr, "Unable to open directory %s", pszDir);
  iRet = SYSERR;
 }

 if (lpszObj)
 {
  FREE(lpszObj);
 }

 return iRet;
}

/**
 * If the passed object is a directory, recurse or
 * ignore it. If it is a file, insert/index it.
 * @param pExtractors Passed to doFile()
 * @param pSock Passed to doFile()
 * @param pszObj The object
 * @returns OK on success, SYSERR on error
 **/

static int processObj(EXTRACTOR_ExtractorList *pExtractors,
                      GNUNET_TCP_SOCKET *pSock, char *pszObj,
                      int bRecurse)
{
 struct stat theStat;
 int iRet;

 if( (stat(pszObj, &theStat) != -1) &&
     (S_ISDIR(theStat.st_mode)) )
 {
  // It's a directory - only recurse if we're allowed to
  if(bRecurse)
  {
   iRet = recurseDir(pExtractors, pSock, pszObj, bRecurse);
  }
  else
  {
   fprintf(stderr, "%s is a directory\n", pszObj);
   iRet = OK;
  }
 }
 else
 {
  // It's a single file
  iRet = dofile(pExtractors, pSock, pszObj);
 }

 return iRet;
}

/**
 * Main method of a tool to insert multiple files into GNUnet.
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return return value from doFile: 0: ok, -1: file not found 
 **/   
int main(int argc, char ** argv) {
  int i;
  int result = 0;
  char ** fileNames;
  int fileNameCount;
  GNUNET_TCP_SOCKET * sock;
  EXTRACTOR_ExtractorList * extractors;
  int pszRecurse;

#if USE_LIBEXTRACTOR == 0
#error "gnunet-insert-multi requires libextractor!"
#endif
  
  if (SYSERR == initUtil(argc, argv, &parseOptions)) 
    return 0;
  fileNameCount = getConfigurationStringList(&fileNames);

  sock = getClientSocket();
  extractors = getExtractors();
  
  pszRecurse = testConfigurationString("GNUNET-INSERT", "RECURSIVE", "YES");

  for (i=0;i<fileNameCount;i++) 
    if (SYSERR == processObj(extractors,
			     sock,
			     fileNames[i],
			     (pszRecurse == YES)))
      result = -1;

  EXTRACTOR_removeAll(extractors);
  releaseClientSocket(sock);
  doneUtil();
  
  return result;
}  

/* end of gnunet-insert-multi.c */
